﻿
-- Add the module to the tree
local mod = klhtm
local me = {}
mod.combat = me

--[[
KTM_Combat.lua

The combat module parses combat log events for damage and abilities done.
]]

-- These are the events we would like to be notified of
me.myevents = { "CHAT_MSG_SPELL_FAILED_LOCALPLAYER", "CHAT_MSG_MONSTER_EMOTE",
"CHAT_MSG_MONSTER_YELL", "CHAT_MSG_COMBAT_FRIENDLY_DEATH", "CHAT_MSG_COMBAT_SELF_HITS",
"CHAT_MSG_SPELL_DAMAGESHIELDS_ON_SELF", "CHAT_MSG_SPELL_SELF_DAMAGE", "CHAT_MSG_SPELL_PERIODIC_CREATURE_DAMAGE",
"CHAT_MSG_SPELL_PERIODIC_PARTY_BUFFS", "CHAT_MSG_SPELL_PERIODIC_SELF_BUFFS", "CHAT_MSG_SPELL_SELF_BUFF",
"CHAT_MSG_SPELL_CREATURE_VS_SELF_DAMAGE" }

-- these are kept for debug purposes. We need two because next attack abilities are split into two.
me.lastattack = nil
me.secondlastattack = nil

--[[ 
This is a record of attacks in the last second while out of combat. We keep this because when you
go into combat by initiating an attack, the +combat event can come after the actual attack. 
]]
me.recentattacks = { } 

me.onupdate = function()

	local timenow = GetTime()
	local key
	local value

	for key, value in me.recentattacks do
		if value[1] < timenow - 1 then
			me.recentattacks[key] = nil
		end
	end

end
	

-- This is a method level temporary variable. Declared at file level because it is a list,
-- and we don't want to keep paging heap memory every time he is created.
me.event = 
{
	["hits"] = 0,
	["damage"] = 0,
	["rage"] = 0,
	["threat"] = 0,
	["name"] = 0,
}

--[[ 
Special onevent() method that will be called by Core.lua:onevent()
]]
me.onevent = function()

	if me.oneventinternal() then
		KLHTM_RequestRedraw("self")
	end
	
end

-- Returns non-nil if the event causes our threat to change
me.oneventinternal = function()
	
	local ability
	local target
	local amount
	local damagetype
	
	if event == "CHAT_MSG_SPELL_FAILED_LOCALPLAYER" then
		if string.find(arg1, mod.string.get("spell", "sunder")) then
			me.retractsundercast()
			return true 
		else
			return
		end
	
	elseif event == "CHAT_MSG_COMBAT_FRIENDLY_DEATH" then
		
		if arg1 == UNITDIESSELF then -- UNITDIESSELF = "You die."
			-- death is a threat wipe
			mod.table.resetraidthreat()
			return true
		end
	
	elseif event == "CHAT_MSG_COMBAT_SELF_HITS" then
		
		_, _, target, amount = string.find(arg1, mod.string.get("combatlog", "whiteattackhit"))
		if target then
			me.registerhit(mod.string.get("threatsource", "whitedamage"), amount, false, target, false)
			return true 
		end

		_, _, target, amount = string.find(arg1, mod.string.get("combatlog", "whiteattackcrit"))
		if target then 
			me.registerhit(mod.string.get("threatsource", "whitedamage"), amount, false, target, true)
			return true
		end

	elseif event == "CHAT_MSG_SPELL_DAMAGESHIELDS_ON_SELF" then
		
		_, _, amount, damagetype, target = string.find(arg1, mod.string.get("combatlog", "damageshield"))
		if amount then 
			
			me.registerhit(mod.string.get("threatsource", "damageshield"), amount, false, target, false, damagetype)
			return true
		end 
		
	elseif event == "CHAT_MSG_SPELL_SELF_DAMAGE" then
			
		-- 1) Spell School Damage Crit
		_, _, ability, target, amount, damagetype = string.find(arg1, mod.string.get("combatlog", "spellcrit"))
		if ability then 
			
			me.registerhit(ability, amount, false, target, true, damagetype)
			return true
		end
		
		-- 2) Spell School Damage Hit
		_, _, ability, target, amount, damagetype = string.find(arg1, mod.string.get("combatlog", "spellhit"))
		if ability then 
			
			me.registerhit(ability, amount, false, target, true, damagetype)
			return true
		end
				
		-- 3) Non-School Ability Hits
		_, _, ability, target, amount = string.find(arg1, mod.string.get("combatlog", "abilityhit"))
		if ability then 
			me.registerhit(ability, amount, false, target, false)
			return true
		end
		
		-- 4) Non-School Ability Crits
		_, _, ability, target, amount = string.find(arg1, mod.string.get("combatlog", "abilitycrit"))
		if ability then 
			me.registerhit(ability, amount, false, target, false)
			return true
		end
		
		-- 5) Special Attacks
		_, _, ability, target = string.find(arg1, mod.string.get("combatlog", "perform"))
		if ability then 
			
			if (mod.string.mylocale == "zhTW") or (mod.string.mylocale == "zhCN") then
				ability, target = target, ability
			end
						
			if ability == mod.string.get("spell", "feint") then
				me.registerhit(ability, 0, false, target, false)
				return true
			
			elseif ability == mod.string.get("spell", "cower") then
				me.registerhit(ability, 0, false, target, false)
				return true
				
			elseif ability == mod.string.get("spell", "taunt") then
				me.registertaunt(target)
				return true
				
			elseif ability == mod.string.get("spell", "growl") then
				me.registertaunt(target)
				return true
			end
			
			return
		end
			
		-- 6) Sunder fails due to parry / dodge / miss
		_, _, target = string.find(arg1, mod.string.get("spell", "sunder"))
		if target then 
			me.retractsundercast(arg1)
			return true
		end
		
		-- 7) Black Amnesty proc
		_, _, ability, target = string.find(arg1, mod.string.get("combatlog", "spellcast"))
		if ability then 
			
			if (mod.string.mylocale == "zhTW") or (mod.string.mylocale == "zhCN") then
				-- {2, 1}
				target, ability = ability, target
			end
			
			if ability == mod.string.get("spell", "reducethreat") then
				me.registerhit("blackamnesty", 0, false, target, false, false)
				return true
				
			elseif ability == mod.string.get("spell", "distractingshot") then
				me.registerhit(ability, 0, false, target, false, false)
				return true 
			end
			
			return
		end

	elseif event == "CHAT_MSG_SPELL_PERIODIC_CREATURE_DAMAGE" then

		_, _, target, amount, damagetype, ability = string.find(arg1, mod.string.get("combatlog", "dot"))
		if target then 
				
		
			if mod.string.mylocale == "frFR" then
				-- 4, 2, 3, 1
				ability, amount, damagetype, target = target, amount, damagetype, ability
				
			elseif (mod.string.mylocale == "zhTW") or (mod.string.mylocale == "zhCN") then
				-- 4, 1, 2, 3
				ability, target, amount, damagetype = target, amount, damagetype, ability
			end
			
			me.registerhit(ability, amount, true, target, false, damagetype)
			
			return true
		end
	
	elseif event == "CHAT_MSG_SPELL_PERIODIC_PARTY_BUFFS" then
		
		-- heal over times deduct overheals automatically
		_, _, target, amount, ability = string.find(arg1, mod.string.get("combatlog", "yourhotonother"))
		if target then 

			if mod.string.mylocale == "frFR" then
				-- 3, 2, 1
				ability, amount, target = target, amount, ability
			end
			
			me.registerpossibleoverheal(ability, amount, target, true)
			
			return true
		end
	
	elseif event ==	"CHAT_MSG_SPELL_PERIODIC_SELF_BUFFS" then
		
		-- a) clean out hots by other players
		 _, _, amount, _, _ = string.find(arg1, mod.string.get("combatlog", "othershotonyou"))
		if amount then 
			return
		end
		
		-- b) hots from us
		_, _, amount, ability = string.find(arg1, mod.string.get("combatlog", "hotonself"))
		if amount then 

			if (mod.string.mylocale == "frFR") or (mod.string.mylocale == "zhTW") or (mod.string.mylocale == "zhCN") then
				-- 2, 1
				amount, ability = ability, amount
			end
			
			me.registerpossibleoverheal(ability, amount, UnitName("player"), true) 
			
			return true
		end
		
		-- c) periodic power gain (bloodrage, essence of the red, etc)
		_, _, amount, ability, source = string.find(arg1, mod.string.get("combatlog", "powergain"))
		if amount then 
			
			if (mod.string.mylocale == "zhTW") or (mod.string.mylocale == "zhCN") then
				-- 3, 1, 2
				source, amount, ability = amount, ability, source
			end
			
			me.registerpowergain(amount, ability, source)
			return true
		end
	
		-- Buff gains
		_, _, ability = string.find(arg1, mod.string.get("combatlog", "buffstart"))
		if ability then 
			
			-- Fetish of the Sand Reaver activation
			if ability == mod.string.get("spell", "arcaneshroud") then
				mod.my.setstate("arcaneshroud", true)
			
			-- Rogue Vanish
			elseif ability == mod.string.get("spell", "vanish") then
				
				-- this is a threat wipe, not a threat reset.
				me.event.hits = 1
				me.event.damage = 0
				me.event.rage = 0
				me.event.threat = - mod.table.getraidthreat()

				me.addattacktodata(mod.string.get("threatsource", "threatwipe"), me.event)
				
				-- threat wipe: add to raid threat offset, not total threat.
				mod.table.raidthreatoffset = mod.table.raidthreatoffset + me.event.threat
				--me.addattacktodata(mod.string.get("threatsource", "total"), me.event)
				
				return true
			
			-- Priest Fade
			elseif ability == mod.string.get("spell", "fade") then
			
				local threat = math.min(mod.table.getraidthreat(), mod.my.ability(mod.string.get("spell", "fade"), "threat"))
				mod.table.raidthreatoffset = mod.table.raidthreatoffset - threat
				mod.my.lastfadevalue = threat
				
				mod.my.setstate("fade", true)
			end
			
			return
		end
				
	elseif event == "CHAT_MSG_SPELL_SELF_BUFF" then

		-- 1) Crit Heal on Self
		_, _, ability, amount = string.find(arg1, mod.string.get("combatlog", "healcritonself"))
		if ability then 
			me.registerpossibleoverheal(ability, amount, UnitName("player"), false)
			return true
		end
		
		-- 2) Non Crit Heal on Self
		_, _, ability, amount = string.find(arg1, mod.string.get("combatlog", "healhitonself"))
		if ability then 
			me.registerpossibleoverheal(ability, amount, UnitName("player"), false)
			return true
		end
		
		-- 3) Crit Heal on Other
		_, _, ability, target, amount = string.find(arg1, mod.string.get("combatlog", "healcritonother"))
		if ability then 
			me.registerpossibleoverheal(ability, amount, target, false)
			return true
		end
		
		-- 4) Heal Hit on Other
		_, _, ability, target, amount = string.find(arg1, mod.string.get("combatlog", "healhitonother"))
		if ability then 
			me.registerpossibleoverheal(ability, amount, target, false)
			return true
		end
		
		-- catch earthfury mana gain, mana pot, etc
		_, _, amount, ability, source = string.find(arg1, mod.string.get("combatlog", "powergain"))
		if ability then 
			
			if mod.string.mylocale == "frFR" then
				-- 3, 1, 2
				source, amount, ability = amount, ability, source
			end
			
			me.registerpowergain(amount, ability, source)
			return true 
		end
		
	elseif event == "CHAT_MSG_SPELL_CREATURE_VS_SELF_DAMAGE" then
		
		_, _, target, ability, amount = string.find(arg1,  mod.string.get("combatlog", "enemyspellhit")) 
		if ability then 
			
			if  mod.string.mylocale == "frFR" then
				-- 2, 1, 3
				ability, target, amount = target, ability, amount
			end
			
			if me.registerpossibleknockback(target, ability) then
				return true
			end
		end
		
	else
		return
	end
	
end

--[[
me.registerpossibleknockback(enemy, ability)
Called when a mob's spell hits you. <enemy> is the name of the mob, <ability> is the spell used.
If this Mob / Ability combination is known to reduce threat, your threat will be reduced.
]]
me.registerpossibleknockback = function(enemy, ability)

	local knockbackdata
	local mobname
	local knockbackvalue
	local mobfound = false

	me.event.hits = 1
	me.event.damage = 0
	me.event.rage = 0
	
	-- Check whether this ability is a known threat reducer
	knockbackdata = mod.data.knockbacks[ability]
	if  knockbackdata == nil then
		return
	end
	
	-- Now look for the specific value for our mob, if it exists
	if knockbackdata[enemy] then
		knockbackvalue = knockbackdata[enemy]
	else
		knockbackvalue = knockbackdata["#Default"]
	end
	
	-- Some knockbacks don't reduce threat (e.g. Onyxia Wing Buffet)
	if knockbackvalue == 0 then 
		return
	end
	
	-- create a combat event and fire it
	me.event.threat = knockbackvalue * mod.table.getraidthreat()
	
	me.addattacktodata(mod.string.get("threatsource", "threatwipe"), me.event)
	
	-- threat wipe: add to raid threat offset, not total threat.
	mod.table.raidthreatoffset = mod.table.raidthreatoffset + me.event.threat
	--me.addattacktodata(mod.string.get("threatsource", "total"), me.event)				
							
	-- Debug Print for Verification:
	mod.out.printtrace("knockback", event)
	
	-- Report it for knockback discovery
	mod.net.sendknockbackevent(enemy, ability, knockbackvalue)
	
end

--[[ 
me.retractsundercast()
Called when an attempted cast of Sunder Armor was found to have failed
]]
me.retractsundercast = function()

	-- 1) check for master target
	if (mod.boss.mastertarget ~= nil) and (UnitName("target") ~= mod.boss.mastertarget) then
		return -- disregard; master target was not targetted
	end

	me.event.hits = -1
	me.event.damage = 0
	me.event.rage = - mod.my.ability(mod.string.get("spell", "sunder"), "rage")
	
	local threatmodifier = mod.my.globalthreat.value
	me.event.threat = - mod.my.ability(mod.string.get("spell", "sunder"), "threat") * threatmodifier 

	me.addattacktodata(mod.string.get("spell", "sunder"), me.event)
	me.addattacktodata(mod.string.get("threatsource", "total"), me.event)
	
	KLHTM_RequestRedraw("self")
end

--[[ 
me.registertaunt()
Called when you succesfully casts Taunt or Growl on a mob.
<target> is the name of the mob you have taunted.
]]
me.registertaunt = function(target)
	
	mod.out.printtrace("targetdebug", string.format("Taunting %s!", target))
	
	--[[
	OK, new idea. If targettarget has greater threat than you, assume they are the aggro target.
	]]
	
	if mod.boss.mastertarget and (target ~= mod.boss.mastertarget) then
		-- you taunted a mob that is not the master target
		return
	end
	
	-- here, you either taunted the master target, or there is no master target
	-- check if you are still targetting that mob (likely, if you just taunted it)
	if UnitName("target") == target then
	
		-- Check for tt
		local targettarget = UnitName("targettarget")
		
		if mod.table.raiddata[targettarget] and (mod.table.raiddata[targettarget] > mod.table.getraidthreat()) then
			-- your current target is targetting another player, and they have more threat than you
			
			local gain = mod.table.raiddata[targettarget] - mod.table.getraidthreat()
				
			me.event.hits = 1
			me.event.damage = 0
			me.event.rage = 0
			me.event.threat = gain
			me.event.name = mod.string.get("spell", "taunt")

			me.addattacktodata(mod.string.get("spell", "taunt"), me.event)
			me.addattacktodata(mod.string.get("threatsource", "total"), me.event)
			
			mod.out.printtrace("targetdebug", string.format("You taunt %s from %s, gaining %d threat.", target, targettarget, gain))
			
			return
		end
	end
	
	-- If that didn't work, use the old code:	
	
	if target == mod.boss.mastertarget then
		local previoustargetthreat = mod.table.raiddata[mod.boss.mttruetarget]
		
		if (previoustargetthreat ~= nil) and (previoustargetthreat > mod.table.getraidthreat()) then
			local tauntgain = previoustargetthreat - mod.table.getraidthreat()
			
			me.event.hits = 1
			me.event.damage = 0
			me.event.rage = 0
			me.event.threat = tauntgain
			me.event.name = mod.string.get("spell", "taunt")
			
			me.addattacktodata(mod.string.get("spell", "taunt"), me.event)
			me.addattacktodata(mod.string.get("threatsource", "total"), me.event)
		
			mod.out.printtrace("targetdebug", string.format("You taunt %s from %s, gaining %d threat.", target, mod.boss.mttruetarget or "<nil>", tauntgain))
		
		else
			mod.out.printtrace("targetdebug", string.format("You taunt %s from %s, but he had a lower threat, so you gain no threat.", target, mod.boss.mttruetarget or "<nil>"))
		end
	end
	
end

--[[
me.registerhit(attack, damage, isdot, target, iscrit, isholy)
Called when most damage abilities are used. Checks your talents and gear and such to get the correct threat.
<attack> is the name of the ability.
<damage> is the damage done.
<isdot> is a boolean, whether the ability was damage over time.
<target> is the name of the mob you hit.
<iscrit> is a boolean, whether the hit was a critical one
<spellschool> is a string, "fire" / "frost" / "arcane" etc, or nil.
]]
me.registerhit = function(attack, damage, isdot, target, iscrit, spellschool)
	
	if damage == nil then
		mod.out.printtrace("assertion", string.format("Damage is nil for attack %s.", tostring(attack)))
		damage = 0
	end
	
	-- Check for Master Target Match
	if mod.boss.mastertarget ~= nil then
		
		-- check for nil target. shoudln't happen, but it is
		if target == nil then
			mod.out.printtrace("assertion", string.format("nil target found. damage = %d, attack = %s.", damage, attack))
		
		elseif target ~= mod.boss.mastertarget then
			-- disregard this completely
			return
		end
	end

	-- threatmodifier includes global things, like defensive stance, tranquil air totem, rogue passive modifier, etc.
	local threatmodifier = mod.my.globalthreat.value
	
	-- Special threat mod: priest silent resolve (spells only)
	if (UnitClass("player") == mod.string.get("class", "priest")) and (attack ~= mod.string.get("threatsource", "whitedamage")) then
		threatmodifier = threatmodifier + mod.my.mods.priest.silentresolve
	end 
	
	-- Special threat modifiers for mages
	if UnitClass("player") == mod.string.get("class", "mage") then
		if spellschool == SPELL_SCHOOL6_CAP then -- arcane
			threatmodifier = threatmodifier + mod.my.mods.mage.arcanethreat
		
		elseif spellschool == SPELL_SCHOOL4_CAP then -- frost
			threatmodifier = threatmodifier + mod.my.mods.mage.frostthreat
		
		elseif spellschool == SPELL_SCHOOL2_CAP then -- fire
			threatmodifier = threatmodifier + mod.my.mods.mage.firethreat
		end
	end
	
	me.event.hits = 1
	me.event.rage = 0
	me.event.name = attack
	me.event.damage = damage
	
	if attack == mod.string.get("threatsource", "whitedamage") then

		-- shaman special: check for rockbiter
		if mod.my.mods.shaman.rockbiter > 0 then
			
			-- make a separate event for the rockbiter
			local weaponspeed = UnitAttackSpeed("player")
			local rockbiterdps = mod.data.rockbiter[mod.my.mods.shaman.rockbiter]

			-- note: we are sending the threat value of rockbiter as the damage argument
			me.registerhit(mod.string.get("spell", "rockbiter"), rockbiterdps * weaponspeed, false, target, iscrit, isholy)
			
			-- now rewrite the correct values to me.event
			me.event.name = attack
			me.event.damage = damage
		end
		
		-- normal behaviour
		me.event.threat = damage * threatmodifier
		
	elseif attack == mod.string.get("spell", "rockbiter") then
		
		me.event.threat = me.event.damage
		me.event.damage = 0

	elseif attack == mod.string.get("spell", "heroicstrike") then
		-- this one is special, hard to integrate with the rest, so it goes on its own
		-- basic idea is to split it into a white damage hit and a heroic strike hit
		
		local preimpaledamage = damage
		if iscrit == true then
			preimpaledamage = damage / (1 + mod.my.mods.warrior.impale)
		end
		
		local addeddamage = mod.my.ability(attack, "nextattack")
		local myaveragedamage = me.averagemainhanddamage()
		local whitedamage = preimpaledamage * (myaveragedamage / (addeddamage + myaveragedamage))
		
		me.registerhit(mod.string.get("threatsource", "whitedamage"), whitedamage, isdot, target, iscrit)
		
		me.event.damage = damage - whitedamage
		me.event.threat = (me.event.damage + mod.my.ability(attack, "threat")) * threatmodifier
		me.event.rage = mod.my.ability(attack, "rage") + whitedamage / (UnitLevel("player") / 2)
		
	elseif attack == mod.string.get("spell", "maul") then
		-- same as heroic strike, but a bit different
		
		local presavagefurydamage = damage / (1 + mod.my.mods.druid.savagefury)		
		local addeddamage = mod.my.ability(attack, "nextattack")
		local myaveragedamage = me.averagemainhanddamage()
		local whitedamage = presavagefurydamage * (myaveragedamage / (addeddamage + myaveragedamage))
		
		me.registerhit(mod.string.get("threatsource", "whitedamage"), whitedamage, isdot, target, iscrit)
		
		me.event.damage = damage - whitedamage
		me.event.threat = threatmodifier * (damage * mod.my.ability(attack, "multiplier") - whitedamage)
		me.event.rage = mod.my.ability(attack, "rage") + whitedamage / (UnitLevel("player") / 2)
		
	elseif attack == mod.string.get("spell", "swipe") then
		-- another special case
		me.event.damage = damage
		me.event.threat = damage * mod.my.ability(attack, "multiplier") * threatmodifier
		me.event.rage = mod.my.ability(attack, "rage") 
	
	elseif (attack == mod.string.get("spell", "feint")) or (attack == mod.string.get("spell", "cower")) then
		
		me.event.threat = mod.my.ability(attack, "threat") * threatmodifier
		
		-- However, feint can't take you negative
		if me.event.threat + mod.table.getraidthreat() < 0 then
			me.event.threat = - mod.table.getraidthreat() 
		end
	
		-- make this attack come up in the "Threat Wipes" row of your personal meter
		attack = mod.string.get("threatsource", "threatwipe")
	
	elseif attack == "blackamnesty" then -- black amnesty proc
		me.event.threat = -540 * threatmodifier
		
		-- However, -threat abilities can't take you negative
		if me.event.threat + mod.table.getraidthreat() < 0 then
			me.event.threat = - mod.table.getraidthreat() 
		end
	
		-- make this attack come up in the "Threat Wipes" row of your personal meter
		attack = mod.string.get("threatsource", "threatwipe")
		
	-- "if there is a row in the personal meter with the same name as this ability, then..."
	elseif mod.table.mydata[attack] then
		
		me.event.threat = (damage + mod.my.ability(attack, "threat")) * threatmodifier
		
		-- check for a multiplier (e.g. searing pain)
		local multi = mod.my.ability(attack, "multiplier")
		if multi then
			me.event.threat = me.event.threat * multi
		end
		
		me.event.rage = mod.my.ability(attack, "rage")

	-- these are periodic_spell_damage
	elseif isdot == true then
		
		me.event.hits = 0
		attack = mod.string.get("threatsource", "dot")
		me.event.threat = damage * threatmodifier
		
	else
		attack = mod.string.get("threatsource", "special")
		me.event.threat = damage * threatmodifier
	end
	
	-- Now apply class-specific filters
	
	-- warlock
	if UnitClass("player") == mod.string.get("class", "warlock") then
			
		-- Nemesis 8/8
		if (mod.my.mods.warlock.nemesis == true) and (mod.data.spellmatchesset("Warlock Destruction", me.event.name) == true) then
			me.event.threat = me.event.threat * 0.8
		end
	
	-- Priest
	elseif UnitClass("player") == mod.string.get("class", "priest") then
			
		-- shadow affinity
		if mod.data.spellmatchesset("Priest Shadow Spells", me.event.name) == true then
			me.event.threat = me.event.threat * mod.my.mods.priest.shadowaffinity
		end
		
	-- Mage
	elseif UnitClass("player") == mod.string.get("class", "mage") then
			
		-- netherwind
		if mod.my.mods.mage.netherwind == true then
			if (me.event.name == mod.string.get("spell", "frostbolt")) or (me.event.name == mod.string.get("spell", "scorch")) then
				me.event.threat = math.max(0, (me.event.threat - 100))
			
			elseif (me.event.name == mod.string.get("spell", "fireball")) and (isdot == false) then
				me.event.threat = math.max(0, (me.event.threat - 100))
				
			elseif me.event.name == mod.string.get("spell", "arcanemissiles") then
				me.event.threat = math.max(0, (me.event.threat - 20))
			end
		end
		
	-- Paladin
	elseif UnitClass("player") == mod.string.get("class", "paladin") then
		
		-- righteous fury
		if spellschool == SPELL_SCHOOL1_CAP then -- holy
			me.event.threat = me.event.threat * mod.my.mods.paladin.righteousfury
		end
	end

	-- Special: abilities that don't cause threat
	if attack == mod.string.get("spell", "holynova") then
		me.event.threat = 0
	end

	-- now add me.event to individual and totals
	me.addattacktodata(attack, me.event)
	
	-- threat wipe: add to raid threat offset, not total threat.
	if attack == mod.string.get("threatsource", "threatwipe") then
		mod.table.raidthreatoffset = mod.table.raidthreatoffset + me.event.threat
	else
		me.addattacktodata(mod.string.get("threatsource", "total"), me.event)
	end
	
end

-- to work out which part of a next attack ability was from white damage
me.averagemainhanddamage = function()

	local min, max = UnitDamage("player")
	return (min + max) / 2

end

--[[
me.registerpossibleoverheal(spell, amount, target, isdot)
Called when you heal someone. Deducts the overhealing from the total, then calls the Heal method
]]
me.registerpossibleoverheal = function(spell, amount, target, isdot)
	
	-- we can check the target's health, which will be the health before the heal. Then we work out what the 
	-- heal did, and we can calculate overheal.
	
	-- for target, 1) check "you", 2) check partyx, 3) check raidx, 4) raidxpet
	
	local unit = me.findunitidfromname(target)
	
	if unit == nil then
		mod.out.printtrace("targetnotinraid", "Could not find a UnitID for the name " .. target)
		-- (and assume there was no overheal)
	else
		local hpvoid = UnitHealthMax(unit) - UnitHealth(unit)
		amount = math.min(amount, hpvoid)
	end
	
	me.registerheal(spell, amount, isdot, target)
	
end

--[[
me.findunitidfromname(name)
Given the name of the target you healed, attempts to find a UnitID from your raidgroup for that target.
We need the target to be in the raid group so that we can find their Max and current HP.
]]
me.findunitidfromname = function(name)
	
	if name == UnitName("player") then
		return "player"
	end 
	
	local x = 0
	
	if GetNumRaidMembers() > 0 then
		for x = 1, 40 do
			if UnitName("raid" .. x) == name then
				return "raid" .. x
			end
		end
		
		for x = 1, 40 do
			if UnitName("raidpet" .. x) == name then
				return "raidpet" .. x
			end
		end
		
	elseif GetNumPartyMembers() > 0 then
		for x = 1, 4 do
			if UnitName("party" .. x) == name then
				return "party" .. x
			end
		end
		
		for x = 1, 4 do
			if UnitName("partypet" .. x) == name then
				return "partypet" .. x
			end
		end
	end
	
end

--[[ 
me.registerheal(spell, amount, isdot, target)
Works out the threat from a heal.
<amount> is the healed amount only (no overheal)
<isdot> is a boolean, whether the spell was a heal over time
<target> is the name of the target
]]
me.registerheal = function(spell, amount, isdot, target)

	if target == mod.string.get("mobs", "ebonroc") then
		return
		
	-- in general, don't count heals towards the master target
	elseif target == mod.boss.mastertarget then
		return
	end

	me.event.name = spell
	me.event.hits = 1
	me.event.rage = 0
	me.event.damage = amount
	
	local threatmod = mod.my.globalthreat.value
	
	-- Special threat mod: priest silent resolve (spells only)
	if UnitClass("player") == mod.string.get("class", "priest") then
		threatmod = threatmod + mod.my.mods.priest.silentresolve 
	end 
	
	me.event.threat = amount * threatmod * mod.data.threatconstants.healing
	
	-- class-based healing multipliers
	if UnitClass("player") == mod.string.get("class", "paladin") then
		me.event.threat = me.event.threat * mod.my.mods.paladin.healing
		
	elseif UnitClass("player") == mod.string.get("class", "druid") then
		me.event.threat = me.event.threat * mod.my.mods.druid.subtlety
		
		if spell == mod.string.get("spell", "tranquility") then
			me.event.threat = me.event.threat * mod.my.mods.druid.subtlety
		end
		
	elseif UnitClass("player") == mod.string.get("class", "shaman") then
		me.event.threat = me.event.threat * mod.my.mods.shaman.healing
		
	end 
	
	-- Special: abilities which don't cause threat
	if spell == mod.string.get("spell", "holynova") then
		me.event.threat = 0
	elseif spell == mod.string.get("spell", "siphonlife") then
		me.event.threat = 0
	elseif spell == mod.string.get("spell", "felstamina") then
		-- me.event.threat = 0
		-- this hasn't been confirmed adequately
		
	elseif spell == mod.string.get("spell", "drainlife") then
		me.event.threat = 0
	end
	
	me.addattacktodata(mod.string.get("threatsource", "healing"), me.event)
	
	me.event.damage = 0
	me.addattacktodata(mod.string.get("threatsource", "total"), me.event)
end

--[[
me.registerpowergain(amount, powertype)
Calculates the threat from gaining energy / mana / rage.
<amount> is the amount of power gained.
<powertype> is "Mana" or "Rage" or "Energy", or it might be the name of a buff, in which case it will be ignored.
This is because we can't differentiate e.g. "You gain 1 rage from Bloodrage" and "You gain 1 Health from Rejuvenation"
until we check.
]]
me.registerpowergain = function(amount, powertype, source)
	
	me.event.name = powertype .. " Gain"
	me.event.damage = amount
	me.event.hits = 1
	me.event.rage = 0
	
	if powertype == mod.string.get("power", "rage") then
		me.event.threat = amount * mod.data.threatconstants.ragegain
		
	elseif powertype == mod.string.get("power", "energy") then
		me.event.threat = amount * mod.data.threatconstants.energygain
		
	elseif powertype == mod.string.get("power", "mana") then
		me.event.threat = amount * mod.data.threatconstants.managain 
		
	else
		return
	end
	
	-- Special: abilities which don't cause threat
	if source == mod.string.get("spell", "felenergy") then
		-- me.event.threat = 0
		-- this hasn't been confirmed adequately
		
	elseif source == mod.string.get("spell", "lifetap") then
		me.event.threat = 0
	end
		
	me.addattacktodata(mod.string.get("threatsource", "powergain"), me.event)
	
	-- now mod it a bit to work into total better
	me.event.damage = 0
	me.event.hits = 0
	
	me.addattacktodata(mod.string.get("threatsource", "total"), me.event)
	
end

--[[
me.addattacktodata(name, data)
Once the threat from an attack has been worked, add it.
<name> is the category to add the threat to
<data> is me.event, i think...
Heap Memory will be created when you call this method when out of combat (mostly healing). 
From the size of the list (two numbers in it), will be 50 bytes tops each time.
]]
me.addattacktodata = function(name, data)
	
	-- Ignore if charmed
	if mod.my.states.playercharmed.value == true then
		return
	end 
	
	if name ~= mod.string.get("threatsource", "total") then
		me.secondlastattack = me.lastattack
		me.lastattack = data
		
		-- Add this to the recent attacks list, if we are not in combat
		if mod.my.states.incombat.value == false then
			table.insert(me.recentattacks, {GetTime(), data.threat})
		end
	end
	
	mod.table.mydata[name].hits = mod.table.mydata[name].hits + data.hits
	mod.table.mydata[name].threat = mod.table.mydata[name].threat + data.threat
	mod.table.mydata[name].rage = mod.table.mydata[name].rage + data.rage
	mod.table.mydata[name].damage = mod.table.mydata[name].damage + data.damage 
	
end


----------------------------------------------------------
--					Handling Sunder						--
----------------------------------------------------------


--[[
klhtu.combat.submitsundercast()
When you think you have just cast a sunder, call this method to make sure the addon knows. There's no way to
reliably detect a sunder hit, so we assume it has been cast, then look for failure signs (misses, spell errors)
It's highly recommended to use klhtm.combat.sunder() or KLHTM_Sunder() instead, because it will make sure
that threat is correct when you spam the button, which may not happen otherwise.
]]
me.submitsundercast = function()
	
	-- check for Master Target
	if (mod.boss.mastertarget ~= nil) and  (UnitName("target") ~= mod.boss.mastertarget) then
		return -- disregard; this is not an action against the master target
	end
	
	me.registerhit(mod.string.get("spell", "sunder"), 0, false, UnitName("target"))
	KLHTM_RequestRedraw("self")
end

-- Call this from a macro, to replace your normal sunder button
me.spellbooksunderindex = 1

-- Kept for compatibility
function KLHTM_Sunder()
	
	me.sunder()
	
end

--[[
me.sunder()
Casts Sunder Armor if most checks reveal it is castable, and calls me.sumbitsundercast().
This method is safe to be spammed. It will only cast if the global cooldown is off, which means once you try to
cast it, you won't try again unless the server says "failed due to <x>". So it's like full good and stuff.
]]
me.sunder = function()
	
	if UnitClass("player") ~= mod.string.get("class", "warrior") then
		return
	end
	
	-- 1) Check if the old position is fine
	if GetSpellName(me.spellbooksunderindex, "spell") ~= mod.string.get("spell", "sunder") then
		
		-- get spell number of spells
		local spelltabs = GetNumSpellTabs()
		local numspells = 0
		local i
		local temp
	
		for i = 1, spelltabs do
			_, _, _, temp = GetSpellTabInfo(i)
			numspells = numspells + temp
		end
	
		-- 2) locate sunder armor in the spell book
		for i = 1, numspells do
			if GetSpellName(me.spellbooksunderindex + i, "spell") == mod.string.get("spell", "sunder") then
				me.spellbooksunderindex = me.spellbooksunderindex + i
				break
				
			elseif i == numspells then
				mod.out.printtrace("sunderdebug", "Can't find sunder in your spellbook!")
				return -- can't find sunder
			end
		end
	end
	
	-- Now we've found sunder. Check the cooldown		
	if GetSpellCooldown(me.spellbooksunderindex, "spell") ~= 0 then 
		return
	end
	
	-- Test for target
	if UnitCanAttack("player", "target") == nil then
		return
	end 
	
	-- Cast
	CastSpellByName(mod.string.get("spell", "sunder") .. "()")
	me.submitsundercast()
	
end

--[[
This code hooks UseAction, intercepts the user casting Sunder Armor, and runs me.sunder() instead.
]]

me.saveduseaction = UseAction

me.newuseaction = function(actionindex, x, y)
	
	-- Check Sunder
	if UnitClass("player") == mod.string.get("class", "warrior") and actionindex ~= nil and GetActionText(actionindex) == nil and GetActionTexture(actionindex) == "Interface\\Icons\\Ability_Warrior_Sunder" then
		if GetActionCooldown(actionindex) > 0 then
			mod.out.printtrace("sunderdebug", "Preventing Sunder cast due to cooldown!")
			return
			
		else
			-- Now check we have a decent target. Otherwise press "tab" - this is what the UI does anyway
			if UnitCanAttack("player", "target") ~= 1 then
				mod.out.printtrace("sunderdebug", "Preventing Sunder cast due to invalid target!")
				return
			end
			
			me.submitsundercast()
			-- At the bottom of this method the sunder will be cast
		end
	
	end

   -- Call the original function
   me.saveduseaction(actionindex, x, y)   

end

UseAction = me.newuseaction
