﻿
--[[

A module can optionally have

1) .myevents 	-	a list of strings, the events to register
2) .onload		-	to be called on startup
3) .onloadcomplete	to be called after all other modules have been loaded.
3) .onevent		-	will do your onevent. Currently sends all events to you, we should fix it to send only your events!
4) .onupdate	-	you can guess this one!
5) .isenabled	-	whether to receive onload, onevent, onupdate commands. Must be <false> to not be called


Managing your Variables / Methods

--> Initialise static data anywhere in your code file
--> Initialise variables that depend on other module's static data in your onload() method
--> Initialise variables that depend on other module's variables in your onloadcomplete() method


Each module has a key / branch in the master table. The following keys have been taken:
	out	
	alert
	console
	string
	combat
	data
	my
	table
	net,	netin
	gui,	guiopt,	guiraid, some other gui
	boss
	diag
]]

-- table setup
klhtm = { }
local me = klhtm
me.frame = nil -- set at runtime 

-- Mod Version
me.release = 16
me.revision = 15
me.build = 192

--[[
Release	Build
	 1	  6
	 2	 11
	 3	 30
	 4	 32
	 5	 44
	 6	 54
	 7	 73
	 8	 80
	 9	 92
	10	103
	11	116
	12	124
	13	141
	14	156
	15	177
	16	189
]]


me.events = { } --[[ 
Remember which module wants which events.
It will look like
{
	["combat"] = 
	{
		["CHAT_MSG_SPELL_SELF_BUFF"] = true,
	},
}
if the combat module has registered CHAT_MSG_SPELL_SELF_BUFF
]]


me.isloaded = false -- true when .onload has been called for all sub-modules
me.isenabled = true -- iif false, onupdate and onevent will not be called

-- onload
me.onload = function()
	
	-- find frame
	me.frame = KLHTM_OnUpdateFrame
	
	-- initialise all submodules
	for key, subtable in me do
		if type(subtable) == "table" and subtable.onload and subtable.isenabled ~= "false" then
			subtable.onload()
		end
	end
	
	me.isloaded = true 
	
	-- register events. Strictly after all modules have been loaded.
	for key, subtable in me do
		if type(subtable) == "table" and subtable.myevents then
			
			me.events[key] = { }
			
			for _, event in subtable.myevents do
				me.frame:RegisterEvent(event)
				me.events[key][event] = true 
			end
		end
	end
	
	-- onloadcomplete
	for key, subtable in me do
		if type(subtable) == "table" and subtable.onloadcomplete and subtable.isenabled ~= "false" then
			subtable.onloadcomplete()
		end
	end
	
	-- Print load message
	me.out.print(string.format(me.string.get("print", "main", "startupmessage"), me.release, me.revision), nil, true)
end

-- OnUpdate
me.onupdate = function()
		
	-- only call when everything has been loaded
	if me.isloaded ~= true then
		return
	end
	
	-- don't call if the entire addon is disabled
	if me.isenabled == false then
		return
	end
	
	for key, subtable in me do
		if type(subtable) == "table" and subtable.onupdate and subtable.isenabled ~= "false" then
			me.diag.logmethodcall(key, "onupdate")
		end
	end
	
end

-- OnEvent
me.onevent = function()

	-- don't call if the entire addon is disabled
	if me.isenabled == false then
		return
	end

	for key, subtable in me do
		-- 1) The subtable is a valid module - is a table and has a .onevent property.
		-- 2) The subtable is not disabled
		-- 3) The subtable has registered the event
		if type(subtable) == "table" and subtable.onevent and subtable.isenabled ~= "false" and me.events[key][event] then
			
			me.diag.logmethodcall(key, "onevent")
		end
	end
	
end


--------------------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------------------
--[[
stupid test thing for table.getn

--> algorithm = removed duplicated entries in a list (keyed by integers)
--> method 1 = use table.getn in for loops and everywhere
--> method 2 = cache length.

can't believe it... it's always 5x, no matter what. not scaling differently. RRRGH!!!

]]
--[[

me.test = function()

	local x
	local list
	local time
	
	for x = 100, 1000, 100 do
		list = me.makelist(x, x / 10)
		time = me.timefunction(me.dupeoutnaive, list)
		
		me.out.print(string.format("took %s ms with %s entries on naive method.", time, x))
	end
	
	for x = 100, 1000, 100 do
		list = me.makelist(x, x / 10)
		time = me.timefunction(me.dupeoutpro, list)
		
		me.out.print(string.format("took %s ms with %s entries on pro method.", time, x))
	end

end


-- returns: a list, length <length>, with values integers from 1 to <range>
me.makelist = function(length, range)

	local list = { }
	local x
	
	for x = 1, length do
		table.insert(list, math.random(range))
	end
	
	return list
	
end

me.timefunction = function(method, arg)

	local start = debugprofilestop()
	method(arg)
	local stop = debugprofilestop()
	
	return stop - start
	
end

me.dupeoutnaive = function(list)

	local x, y, value
	
	for x = 1, table.getn(list) do
		value = list[x]
		
		y = x + 1
		
		while y <= table.getn(list) do
			if list[y] == value then
				list[y] = list[table.getn(list)]
				table.remove(list, table.getn(list))
			else
				y = y + 1
			end
		end
	end

end

me.dupeoutpro = function(list)

	local x, y, value, count
	
	count = table.getn(list)
	
	for x = 1, count do
		value = list[x]
		
		y = x + 1
		
		while y <= count do
			if list[y] == value then
				list[y] = list[count]
				table.remove(list, count)
				count = count - 1
			else
				y = y + 1
			end
		end
	end

end

]]