﻿
-- Add the module to the tree
local mod = klhtm
local me = {}
mod.data = me

--[[ 
KTM_Data

A list of constants, and a few helper methods.

]]--

-- these are initialised at onload(), below
me.knockbacks = { }
me.abilities = { } 

--[[
Special onload() method called by Core. We have to do this stuff in onload() to be sure the localisation
module has been initialised.
]]
me.onload = function()
	
	me.definespellsets()
	me.definetalents()
	me.defineitemsets()
	
	-- These are the amounts you lose, as a fraction. Should all be negative.
	me.knockbacks =
	{	
		[mod.string.get("knockbacks", "knockaway")] = 
		{
			["#Default"] = -0.5,
			[mod.string.get("mobs", "onyxia")] = -0.333,
		},
		[mod.string.get("knockbacks", "wingbuffet")] = 
		{
			["#Default"] = -0.5,
			[mod.string.get("mobs", "onyxia")] = 0,
		},
	}
	
	me.abilities = 
	{	
		[mod.string.get("spell", "heroicstrike")] = 
		{	
			["default"] = 
			{
				["threat"] = 145,
				["nextattack"] = 138,
				["rage"] = 15,
			},
			["9"] = 
			{
				["threat"] = 175,
				["nextattack"] = 157,
			}
		},
		[mod.string.get("spell", "maul")] = 
		{	
			["default"] = 
			{
				["threat"] = 0,
				["nextattack"] = 128,
				["rage"] = 15,
				["multiplier"] = 1.75,
			}
		},
		[mod.string.get("spell", "swipe")] = 
		{	
			["default"] = 
			{
				["threat"] = 0,
				["rage"] = 20,
				["multiplier"] = 1.75,
			},
		},
		[mod.string.get("spell", "shieldslam")] = 
		{	
			["default"] = 
			{
				["threat"] = 250,
				["rage"] = 20,
			}
		},
		[mod.string.get("spell", "revenge")] = 
		{	
			["default"] = 
			{
				["threat"] = 315,
				["rage"] = 5,
			},
			["6"] = 
			{
				["threat"] = 355,
			},
		},
		[mod.string.get("spell", "shieldbash")] = 
		{	
			["default"] = 
			{
				["threat"] = 180,
				["rage"] = 10,
			},
		},
		[mod.string.get("spell", "sunder")] = 
		{	
			["default"] = 
			{
				["threat"] = 260,
				["rage"] = 15,
			},
		},
		[mod.string.get("spell", "feint")] = 
		{
			["default"] = 
			{
				["threat"] = -600,
				["rage"] = 0,
			},
			["5"] = 
			{
				["threat"] = -800
			}
		},
		[mod.string.get("spell", "cower")] = 
		{
			["default"] = 
			{
				["threat"] = -600,
				["rage"] = 0,
			},
		},
		[mod.string.get("spell", "searingpain")] = 
		{	
			["default"] = 
			{
				["multiplier"] = 2.0,
				["threat"] = 0,
				["rage"] = 0,
			},
		},
		[mod.string.get("spell", "earthshock")] = 
		{	
			["default"] = 
			{
				["multiplier"] = 2.0,
				["threat"] = 0,
				["rage"] = 0,
			},
		},
		[mod.string.get("spell", "mindblast")] = 
		{	
			["default"] = 
			{
				["multiplier"] = 2.0,
				["threat"] = 0,
				["rage"] = 0,
			},
		},
		[mod.string.get("spell", "holyshield")] = 
		{	
			["default"] = 
			{
				["multiplier"] = 1.2,
				["threat"] = 0,
				["rage"] = 0,
			},
		},
		[mod.string.get("spell", "distractingshot")] = 
		{
			["default"] = 
			{
				["threat"] = 600,
				["rage"] = 0,
			},
			["1"] = 
			{
				["threat"] = 110
			},
			["2"] = 
			{
				["threat"] = 160
			},
			["3"] = 
			{
				["threat"] = 250
			},
			["4"] = 
			{
				["threat"] = 350
			},
			["5"] = 
			{
				["threat"] = 465
			},
			["6"] = 
			{
				["threat"] = 600
			},			
		},
		[mod.string.get("spell", "fade")] = 
		{
			["default"] = 
			{
				["threat"] = 820,
				["rage"] = 0,
			},
			["1"] = 
			{
				["threat"] = 55
			},
			["2"] = 
			{
				["threat"] = 155
			},
			["3"] = 
			{
				["threat"] = 285
			},
			["4"] = 
			{
				["threat"] = 440
			},
			["5"] = 
			{
				["threat"] = 620
			},
			["6"] = 
			{
				["threat"] = 820
			},			
		},
		[mod.string.get("spell", "thunderfury")] = 
		{
			["default"] = 
			{
				["threat"] = 241,
				["rage"] = 0,
			},
		},
	}

end

-- These are the DPS modifiers for ranks of rockbiter. Whenever a hit lands with a rockbiter weapon, the added threat
-- equals the speed of the weapon times the rockbiter value, e.g. 72 dps for max rank.
me.rockbiter = 
{
	[1] = 6,  --  1
	[2] = 10, --  8
	[3] = 16, -- 16
	[4] = 27, -- 24
	[5] = 41, -- 34
	[6] = 55, -- 44
	[7] = 72, -- 54
}

-- A bunch of firm constants. 
me.threatconstants = 
{	
	["healing"] = 0.5,
	["meleeaggrogain"] = 1.1,
	["rangeaggrogain"] = 1.3,
	["ragegain"] = 5.0,
	["energygain"] = 2.5,
	["managain"] = 0.5,	
}

--[[ 
mod.data.isbuffpresent(texture)
Looks in your buff list for an icon matching the supplied texture.
<texture> is the path of the texture, e.g. "Interface\\Icons\\Ability_Warrior_Sunder"
Returns: true if the buff is present, false otherwise
]]
me.isbuffpresent = function(texture)
	
	local x
	local bufftexture
	
	for x = 1, 16 do
		bufftexture = UnitBuff("player", x)
		
		if bufftexture == nil then
			break
			
		elseif bufftexture == texture then
			return true
		end
	end
			
	return false
end

--------------------------------------------------------------------------

------------------------------
--        Spell Sets        --
------------------------------

--[[ 
Certain items and abilities only affect particular schools of spells. For these specific sets, we keep
a list of all the possible spells, and provide a method to query a spell as from a school.
]]
me.definespellsets = function()

me.spellsets = 
{
	["Warlock Destruction"] = 
	{
		mod.string.get("spell", "shadowbolt"),
		mod.string.get("spell", "immolate"),
		mod.string.get("spell", "conflagrate"), 
		mod.string.get("spell", "searingpain"), 
		mod.string.get("spell", "rainoffire"), 
		mod.string.get("spell", "soulfire"), 
		mod.string.get("spell", "shadowburn"), 
		mod.string.get("spell", "hellfire"), 
	},
	["Mage Offensive Arcane"] = 
	{
		mod.string.get("spell", "arcanemissiles"), 
		mod.string.get("spell", "arcaneexplosion"), 
		mod.string.get("spell", "counterspell"), 
	},
	["Priest Shadow Spells"] = 
	{
		mod.string.get("spell", "mindflay"), 
		mod.string.get("spell", "devouringplague"), 
		mod.string.get("spell", "shadowwordpain"), 
		mod.string.get("spell", "mindblast"), 
		mod.string.get("spell", "manaburn"), 
	}
}

end
--[[ 
mod.data.spellmatchesset(setname, spellname)
Returns: true if the spell is in the set, false otherwise.
<setname> is a key to me.spellsets above, e.g. "Priest Shadow Spells".
<spellname> is the name of a spell. It is localised.
]]
me.spellmatchesset = function(setname, spellname)
	
	local x = 0
	local spellset = me.spellsets[setname]
	local spell
	
	while true do
		x = x + 1
		spell = spellset[x]
		
		if spell == nil then 
			return false
		
		elseif spell == spellname then 
			return true
		end
	end
	
end


--------------------------------------------------------------------------

------------------------------
--      Talent Points       --
------------------------------

me.definetalents = function()

-- Values are {Page, Talent, Class}
me.talentinfo = 
{	
	[mod.string.get("spell", "sunder")] = {3, 10, mod.string.get("class", "warrior")},
	[mod.string.get("spell", "heroicstrike")] = {1, 1, mod.string.get("class", "warrior")},
	[mod.string.get("talent", "defiance")] = {3, 9, mod.string.get("class", "warrior")},
	[mod.string.get("talent", "impale")] = {1, 11, mod.string.get("class", "warrior")},
	[mod.string.get("talent", "silentresolve")] = {1, 3, mod.string.get("class", "priest")},
	[mod.string.get("talent", "shadowaffinity")] = {3, 3, mod.string.get("class", "priest")},
	[mod.string.get("talent", "druidsubtlety")] = {3, 8, mod.string.get("class", "druid")},
	[mod.string.get("talent", "feralinstinct")] = {2, 3, mod.string.get("class", "druid")},
	[mod.string.get("talent", "ferocity")] = {2, 1, mod.string.get("class", "druid")},
	[mod.string.get("talent", "tranquility")] = {3, 13, mod.string.get("class", "druid")},
	[mod.string.get("talent", "savagefury")] = {2, 13, mod.string.get("class", "druid")},
	[mod.string.get("talent", "masterdemonologist")] = {2, 15, mod.string.get("class", "warlock")},
	[mod.string.get("talent", "arcanesubtlety")] = {1, 1, mod.string.get("class", "mage")},
	[mod.string.get("talent", "frostchanneling")] = {3, 12, mod.string.get("class", "mage")},
	[mod.string.get("talent", "burningsoul")] = {2, 9, mod.string.get("class", "mage")},
	[mod.string.get("talent", "righteousfury")] = {2, 7, mod.string.get("class", "paladin")},
	[mod.string.get("talent", "healinggrace")] = {3, 9, mod.string.get("class", "shaman")},
}

end

--[[ 
me.gettalentrank(talent)
Returns: how many points you have invested in the specified talent.
<talent> is a value from the me.talentinfo array.
]]
me.gettalentrank = function(talent)
	
	local info = me.talentinfo[talent]
	local rank
	_, _, _, _, rank = GetTalentInfo(info[1], info[2])
	
	return rank
end

-- This is a pretty simple function to print out the talents you havee that the mod is checking for
me.testtalents = function()
	local myclass = UnitClass("player")
	
	local key
	local value
	local rank
	local numtalents = 0
	
	for key, value in me.talentinfo do
		if value[3] == myclass then
			rank = me.gettalentrank(key)
			mod.out.print(string.format(mod.string.get("print", "data", "talentpoint"), rank, key))
			numtalents = numtalents + 1
		end
	end
	
	mod.out.print(string.format(mod.string.get("print", "data", "talent"), numtalents, myclass))
end


--------------------------------------------------------------------------

-------------------------------------------
--        Checking for Set Pieces        --
-------------------------------------------

-- This will print a list of all the (significant) set pieces you are wearing.
me.testitemsets = function()
	
	local setname
	local output
	local pieces
	
	for setname in me.itemsets do
		output = setname .. ": {" 
		_, pieces = me.getsetpieces(setname, "non-nil")
		output = output .. pieces .. "}"
		mod.out.print(output)
	end
	
end

-- the key is the description, the value is the item slot index, for GetInventoryItemLink
me.itemslots = 
{
	["head"] = 1,
	["legs"] = 7,
	["shoulder"] = 3,
	["feet"] = 8,
	["waist"] = 6,
	["wrist"] = 9, 
	["chest"] = 5,
	["hands"] = 10,
}

me.defineitemsets = function()

-- values are item numbers. The numbers will be contained in an item link string.
me.itemsets = 
{
	[mod.string.get("sets", "might")] = 
	{
		["head"] = "16866",
		["legs"] = "16867",
		["shoulder"]= "16868",
		["feet"] = "16862",
		["waist"] = "16864",
		["wrist"] = "16861",
		["chest"] = "16865",
		["hands"] = "16863",
	},
	[mod.string.get("sets", "bloodfang")] = 
	{
		["head"] = "16908",
		["legs"] = "16909",
		["shoulder"]= "16832",
		["feet"] = "16906",
		["waist"] = "16910",
		["wrist"] = "16911",
		["chest"] = "16905",
		["hands"] = "16907",
	},
	[mod.string.get("sets", "arcanist")] = 
	{
		["head"] = "16795",
		["legs"] = "16796",
		["shoulder"]= "16797",
		["feet"] = "16800",
		["waist"] = "16802",
		["wrist"] = "16799",
		["chest"] = "16798",
		["hands"] = "16801",
	},
	[mod.string.get("sets", "netherwind")] = 
	{
		["head"] = "16914",
		["legs"] = "16915",
		["shoulder"]= "16917",
		["feet"] = "16912",
		["waist"] = "16818",
		["wrist"] = "16918",
		["chest"] = "16916",
		["hands"] = "16913",
	},
	[mod.string.get("sets", "nemesis")] = 
	{
		["head"] = "16929",
		["legs"] = "16930",
		["shoulder"]= "16932",
		["feet"] = "16927",
		["waist"] = "16933",
		["wrist"] = "16934",
		["chest"] = "16931",
		["hands"] = "16928",
	},
}

end

local setlist
--[[ 
mod.data.getsetpieces(setname, isdebug)
Returns: the number of set pieces the player is currently wearing.
<setname> is the localised name of the set.
if <isdebug> is true, the method will also generate and return as the second value a printout.
]]
me.getsetpieces = function(setname, isdebug)
	
	-- 1) Get the set list
	setlist = me.itemsets[setname]
	
	if setlist == nil then
		me.out.printtrace("assertion", string.format("The set |cffffff00%s|r does not exist in our database.", setname))
		return 0
	end
	
	local slotname
	local slotnumber
	local debugout = ""
	local itemlink
	local numitems = 0
	
	for slotname, slotnumber in me.itemslots do
		itemlink = GetInventoryItemLink("player", slotnumber)
		
		if itemlink and string.find(itemlink, setlist[slotname]) then
			numitems = numitems + 1
			
			-- if it's for debug, print out which piece it is
			if isdebug then
				if numitems > 1 then
					debugout = debugout .. ", "
				end
				
				debugout = debugout .. slotname
			end
		end
	end
	
	return numitems, debugout
	
end
